import itertools
import logging
import math
from copy import copy
from functools import partial
from typing import Generator, Iterable, Optional, Set, List

from tqdm import tqdm

from alparc.types.base_types import Register, RegisterType
from alparc.types.phoneme import PHONEME_FEATURE_LABELS, TypePhonemeFeatureLabels
from alparc.types.syllable import SyllableType
from alparc.types.word import WordType, Word
from alparc.types.lexicon import Lexicon, LexiconType

from alparc.core.word import word_overlap_matrix

from alparc.controls.common import *

logger = logging.getLogger(__name__)

def word_as_syllable_set(word: WordType) -> Set[SyllableType]:
    return set(syllable.id for syllable in word)


def check_no_syllable_overlap(words: RegisterType, word_index_pair):
    index_word_1, index_word_2 = set(word_index_pair)

    syllables_word_1 = word_as_syllable_set(words[index_word_1])
    syllables_word_2 = word_as_syllable_set(words[index_word_2])

    intersection = syllables_word_1.intersection(syllables_word_2)

    return not intersection


def make_lexicon_generator(
        words: RegisterType,
        n_words: int = 4,
        max_overlap: int = 1,
        max_yields: int = 1_000_000,
        lag_of_interest: int = 1,
        control_features: List[TypePhonemeFeatureLabels] = PHONEME_FEATURE_LABELS,) -> Generator[Lexicon, None, None]:

    overlap = word_overlap_matrix(words, lag_of_interest=lag_of_interest, control_features=control_features)
    options = dict((k, v) for k, v in locals().items() if not k == 'words' and not k == 'overlap')
    logger.info(f"GENERATE MIN OVERLAP LEXICONS WITH OPTIONS {options}")
    yields = 0

    iter_allowed_overlaps = itertools.product(range(max_overlap + 1), range(1, math.comb(n_words, 2)))

    for max_pair_overlap, max_overlap_with_n_words in iter_allowed_overlaps:

        max_cum_overlap = max_pair_overlap * max_overlap_with_n_words

        if max_pair_overlap != 0:
            logger.warning(
                f"Increasing allowed overlaps: "
                f"MAX_PAIRWISE_OVERLAP={max_pair_overlap}, "
                f"MAX_CUMULATIVE_OVERLAP={max_cum_overlap}"
            )

        # WORDSxWORDS boolean matrix indicating if the words can be paired together based on the maximum overlap
        valid_word_pairs_matrix = (overlap <= max_pair_overlap)

        # represent the matrix from above as a set of pairs of word indexes, e.g. {{0, 1}, {0, 2}, ...}
        min_overlap_pairs = zip(*np.where(valid_word_pairs_matrix))
        min_overlap_pairs = set(frozenset([int(pair[0]), int(pair[1])]) for pair in min_overlap_pairs)

        # select only those pairs that have pairwise unique syllables
        no_syllable_overlap = partial(check_no_syllable_overlap, words)
        min_overlap_pairs = set(filter(no_syllable_overlap, min_overlap_pairs))

        for start_pair in min_overlap_pairs:

            lexicon_indexes = set(start_pair)
            cumulative_overlap = 0

            for candidate_idx in range(len(overlap)):

                # is candidate word known?
                if candidate_idx in lexicon_indexes:
                    continue

                # does the new word exceed the allowed pairwise overlap?
                has_min_overlap = [({known_idx, candidate_idx} in min_overlap_pairs) for known_idx in lexicon_indexes]
                if not all(has_min_overlap):
                    continue

                # make sure the new word does not increase the cumulative overlap too much
                sum_overlaps_with_known = sum([overlap[known_idx, candidate_idx] for known_idx in lexicon_indexes])
                current_overlap_budget = (max_cum_overlap - cumulative_overlap)
                if sum_overlaps_with_known > current_overlap_budget:
                    continue

                # success! let's add the new word
                lexicon_indexes.add(candidate_idx)
                cumulative_overlap += sum_overlaps_with_known

                # do we have to go again?
                if len(lexicon_indexes) < n_words:
                    continue

                # yield lexicon (guaranteed to be the next best)
                lexicon = Register({words[idx].id:  words[idx] for idx in lexicon_indexes})
                lexicon.info = copy(words.info)
                lexicon.info["cumulative_feature_repetitiveness"] = int(cumulative_overlap)
                lexicon.info["max_pairwise_feature_repetitiveness"] = int(max_pair_overlap)
                yield lexicon

                yields += 1

                if yields == max_yields:
                    return


def sample_random_lexicon(
    words: RegisterType,
    n_words: int = 4
) -> Generator[Lexicon, None, None]:
    while True:
        random_word_indexes = random.sample(range(len(words)), n_words)
        lexicon = Register({words[idx].id: words[idx] for idx in random_word_indexes})
        lexicon.info = copy(words.info)

        sylls = [syll.id for word in lexicon for syll in word]
        intersection = (len(set(sylls)) == len(sylls))

        if intersection:
            continue

        yield lexicon


def make_lexicons(
    words: RegisterType,
    n_lexicons: int = 5,
    n_words: int = 4,
    max_overlap: int = 1,
    lag_of_interest: int = 1,
    max_word_matrix: int = 200,
    unique_words: bool = False,
    binary_feature_control: bool = True,
    progress_bar: bool = False,
    control_features: List[TypePhonemeFeatureLabels] = PHONEME_FEATURE_LABELS,
) -> List[Lexicon]:
    """_summary_

    Args:
        words (RegisterType): The Register of words which the lexicon generation is based on.
        n_lexicons (int, optional): How many lexicons to generate. Defaults to 5.
        n_words (int, optional): How many words should be in a lexicon. Defaults to 4.
        max_overlap (int, optional): How much feature overlap between pairwise word features is allowed. Defaults to 1.
        lag_of_interest (int, optional): the frequency of the word features for which a feature is consideret 'overlapping'. 1 means the feature frequency is the number of syllables in 1 word. Defaults to 1.
        max_word_matrix (int, optional): How many words to use maximum (subsample if nessesary) to generate the feature overlap matrix. Defaults to 200.
        unique_words (bool, optional): check uniqueness of words across all lexicons. Defaults to False.
        binary_feature_control (bool): control feature overlap between words in the lexicon. If false lexicons will be generated completely at random. Defaults to True.

    Returns:
        List[Lexicon]: A List of Lexicons
    """

    lexicons = []

    if progress_bar:
        pbar = tqdm(total=n_lexicons)

    if binary_feature_control:
        lexicon_generator = make_lexicon_generator(
            words.get_subset(max_word_matrix),
            n_words=n_words,
            max_overlap=max_overlap,
            lag_of_interest=lag_of_interest,
            control_features=control_features,
        )
    else:
        lexicon_generator: Iterable = sample_random_lexicon(
            words.get_subset(max_word_matrix),
            n_words=n_words,
        )

    for lexicon in lexicon_generator:

        has_repeating_words = False
        if unique_words:
            # check uniqueness of words across all lexicons
            for lexicon_known in lexicons:
                if set(lexicon_known.keys()).intersection(set(lexicon.keys())):
                    has_repeating_words = True
                    break

        if not has_repeating_words:
            lexicons.append(lexicon)
            if progress_bar:
                pbar.update(1)

        if len(lexicons) >= n_lexicons:
            break
    
    if not binary_feature_control:
        for lexicon in lexicons:
            overlap = word_overlap_matrix(lexicon)
            lexicon.info["cumulative_feature_repetitiveness"] = int(np.triu(overlap, 1).sum())
            lexicon.info["max_pairwise_feature_repetitiveness"] = int(np.triu(overlap, 1).max())

    return lexicons



from typing import TypeVar, List, Dict, Any

from pydantic import BaseModel

from alparc.types.base_types import Register, Element, RegisterType
from alparc.types.syllable import Syllable, SyllableType


WordType = TypeVar("WordType", bound="Word")


class Word(Element, BaseModel):
    id: str
    syllables: List[SyllableType]
    info: Dict[str, Any]

    def get_elements(self):
        return self.syllables



import itertools
import logging
from copy import copy
from functools import reduce
from typing import List, Literal, Optional, Union, TypeVar, Dict, Any

from pydantic import BaseModel

from alparc.types.base_types import Register, RegisterType
from alparc.types.phoneme import Phoneme, TypePhonemeFeatureLabels
from alparc.types.base_types import Element
from alparc.types.syllable import Syllable, SyllableType, LABELS_C, LABELS_V

from alparc.io import read_syllables_corpus

from alparc.controls.filter import filter_common_phoneme_syllables, filter_uniform_syllables

logger = logging.getLogger(__name__)

def add_phonotactic_features(syllable_phonemes: List[Phoneme]):
    syll_feats = [[] for _ in syllable_phonemes]
    for i, phon in enumerate(syllable_phonemes):
        is_consonant = phon.get_binary_feature("cons")
        if is_consonant:
            if phon.get_binary_feature("son"):
                syll_feats[i].append("son")
            if not phon.get_binary_feature("son") and not phon.get_binary_feature("cont"):
                syll_feats[i].append("plo")
            if not phon.get_binary_feature("son") and phon.get_binary_feature("cont"):
                syll_feats[i].append("fri")
            if phon.get_binary_feature("lab"):
                syll_feats[i].append("lab")
            if phon.get_binary_feature("cor") and not phon.get_binary_feature("hi"):
                syll_feats[i].append("den")
            if "lab" not in syll_feats[i] and "den" not in syll_feats[i]:
                syll_feats[i].append("oth")
        else:
            for vowel in ['a', 'e', 'i', 'o', 'u', 'ɛ', 'ø', 'y']:
                if vowel in phon.id:
                    syll_feats[i].append(vowel)

    return syll_feats


def get_feature_labels(phoneme: Phoneme):
    if phoneme.get_binary_feature('cons'):
        return LABELS_C
    return LABELS_V


def syllable_from_phonemes(phonemes: RegisterType, phoneme_combination: List[str], syll_feature_labels: List[List[str]] = None):
    syll_phons = []
    syll_features = []
    for p in phoneme_combination:
        phoneme: Phoneme = phonemes[p]
        syll_phons.append(phoneme)
        for label in get_feature_labels(phoneme):
            if phoneme.get_binary_feature(label):
                syll_features.append(1)
            else:
                syll_features.append(0)

    syllable = Syllable(
        id="".join(phoneme_combination), 
        info={"binary_features": syll_features,
                          "phonotactic_features": add_phonotactic_features(syll_phons)},
        phonemes=syll_phons
    )

    return syllable


def make_feature_syllables(
    phonemes: RegisterType,
    phoneme_pattern: Union[str, list] = "cV",
    max_combinations: int = 1_000_000,
    consonant_features: List[TypePhonemeFeatureLabels] = LABELS_C,
    vowel_features: List[TypePhonemeFeatureLabels] = LABELS_V,
) -> RegisterType:
    """Generate syllables form feature-phonemes. Only keep syllables that follow the phoneme pattern"""

    logger.info("SELECT SYLLABLES WITH GIVEN PHONEME-TYPE PATTERN AND WITH PHONEMES WE HAVE FEATURES FOR")
    valid_phoneme_types = ["c", "C", "v", "V"]

    phoneme_types_user = list(phoneme_pattern) if isinstance(phoneme_pattern, str) else phoneme_pattern

    phoneme_types = list(filter(lambda p: p in valid_phoneme_types, phoneme_types_user))

    if phoneme_types_user != phoneme_types:
        logger.warning(f"ignoring invalid phoneme types {phoneme_types_user} -> {phoneme_types}. "
              f"You can use the following phoneme types in your pattern: {valid_phoneme_types}")

    logger.info(f"Search for phoneme-pattern '{''.join(phoneme_types)}'")

    labels_mapping = {"c": consonant_features, "C": consonant_features, "v": vowel_features, "V": vowel_features}
    syll_feature_labels = list(map(lambda t: labels_mapping[t], phoneme_types))

    single_consonants, multi_consonants, short_vowels, long_vowels = [], [], [], []
    for phoneme in phonemes.values():
        if phoneme.get_binary_feature('cons'):
            if len(phoneme.id) == 1:
                single_consonants.append(phoneme.id)
            else:
                multi_consonants.append(phoneme.id)
        else:
            if len(phoneme.id) == 2 and phoneme.get_binary_feature('long'):
                    long_vowels.append(phoneme.id)
            if len(phoneme.id) == 1 and not phoneme.get_binary_feature('long'):
                    short_vowels.append(phoneme.id)

    phonemes_mapping = {"c": single_consonants, "C": multi_consonants, "v": short_vowels, "V": long_vowels}

    phonemes_factors = list(map(lambda phoneme_type: phonemes_mapping[phoneme_type], phoneme_types))
    total_combs = reduce(lambda a, b: a * b, [len(phonemes_factor) for phonemes_factor in phonemes_factors])
    if total_combs > max_combinations:
        logger.warning(f"Combinatorial explosion with {total_combs} combinations for '{phoneme_types}'."
                        f"I will only generate {max_combinations} of them, but you can set this number higher via the "
                        "option 'max_combinations'.")

    syllables_dict = {}
    list_of_combinations = []
    for i, phoneme_combination in enumerate(itertools.product(*phonemes_factors)):
        if i >= max_combinations:
            break
        list_of_combinations.append(phoneme_combination)

    for phoneme_combination in list_of_combinations:
        syllable = syllable_from_phonemes(phonemes, phoneme_combination, syll_feature_labels)
        syllables_dict[syllable.id] = syllable

    new_info = copy(phonemes.info)
    new_info.update({"syllable_feature_labels": syll_feature_labels,  "syllable_type": phoneme_pattern})

    return Register(syllables_dict, _info=new_info)


def make_syllables(phonemes: RegisterType, phoneme_pattern: str = "cV",
                   #unigram_control: bool = True,
                   syllable_control: bool = True, 
                   syllable_alpha: Optional[float] = 0.05,
                   lang: str = "deu",
                   consonant_features: List[TypePhonemeFeatureLabels] = LABELS_C,
                   vowel_features: List[TypePhonemeFeatureLabels] = LABELS_V,) -> RegisterType:
    """_summary_

    Args:
        phonemes (RegisterType): A Register of phonemes that will be used as a basis to generate the syllables
        phoneme_pattern (str, optional): describes how a syllable is structured, e.g. "cV" syllables consist of a single-consonant character and a long vowel. Defaults to "cV".
        unigram_control (bool, optional): apply statistical control (on the basis of p-val of uniform distribution) to single unigrams. Defaults to True.
        language_control (bool, optional): apply language specific controls (only german for now) on the syllable level. Defaults to True.
        language_alpha (Optional[float], optional): which p-value to assume for language based statistical control. Defaults to 0.05.
        from_format (Literal[&quot;ipa&quot;, &quot;xsampa&quot;], optional): language control will read from a syllable corpus. which format to assume. Defaults to "xsampa".
        lang (str, optional): which language to use for language controls. Defaults to "deu".

    Returns:
        RegisterType: The final Register of syllables
    """

    syllables = make_feature_syllables(phonemes, 
                                       phoneme_pattern=phoneme_pattern,
                                       consonant_features=consonant_features,
                                       vowel_features=vowel_features)

    if syllable_control:
        syllable_corpus = read_syllables_corpus(lang=lang)
        syllables = syllables.intersection(syllable_corpus)
    
        if syllable_alpha is not None:
            syllables = filter_uniform_syllables(syllables, alpha=syllable_alpha)
    
    #if unigram_control and lang == "deu":
    #    syllables = filter_common_phoneme_syllables(syllables)

    return syllables



from .base_types import Register, RegisterType, Element
from .phoneme import Phoneme, PhonemeType
from .syllable import Syllable, SyllableType
from .word import Word, WordType
from .lexicon import Lexicon, LexiconType
from .stream import Stream, StreamType



import json
from abc import ABC, abstractmethod
from os import PathLike

from typing import List, Literal, get_args, Dict, Any, Union

from pydantic import BaseModel

TypePhonemeFeatureLabels = Literal[
    "syl", "son", "cons", "cont", "delrel", "lat", "nas", "strid", "voi", "sg", "cg", "ant", "cor", "distr", "lab",
    "hi", "lo", "back", "round", "tense", "long"
]
PHONEME_FEATURE_LABELS = list(get_args(TypePhonemeFeatureLabels))

LABELS_C = ['son', 'back', 'hi', 'lab', 'cor', 'cont', 'lat', 'nas', 'voi']
LABELS_V = ['back', 'hi', 'lo', 'lab', 'tense', 'long']
N_FEAT = len(LABELS_C) + len(LABELS_V)  # 14

class Element(ABC):
    id: str
    info: Dict[str, Any]

    def __getitem__(self, item):
        return self.get_elements()[item]

    def __iter__(self):
        return iter(self.get_elements())

    def __str__(self):
        return self.id

    @abstractmethod
    def get_elements(self):
        pass


class Phoneme(Element, BaseModel):
    id: str
    info: Dict[str, Any]

    def get_elements(self):
        return []

    def get_feature_symbol(self, label: TypePhonemeFeatureLabels):
        return self.info["features"][PHONEME_FEATURE_LABELS.index(label)]

    def get_binary_feature(self, label: TypePhonemeFeatureLabels):
        return self.get_feature_symbol(label) == "+"


class Syllable(Element, BaseModel):
    id: str
    phonemes: List[Phoneme]
    info: Dict[str, Any]

    def get_elements(self):
        return self.phonemes


class Word(Element, BaseModel):
    id: str
    syllables: List[Syllable]
    info: Dict[str, Any]

    def get_elements(self):
        return self.syllables


class Stream(Element, BaseModel):
    id: str
    syllables: List[Syllable]
    info: Dict[str, Any]

    def get_elements(self):
        return self.syllables
        
    def __str__(self):
        return "|".join(syllable.id for syllable in self)
    
    def save(self, path: Union[str, PathLike] = None):
        if path is None:
            path = f"stream.json"

        if isinstance(path, str) and not path.endswith(".json"):
            path = path + ".json"

        with open(path, "w", encoding='utf-8') as file:
            json.dump(self, file,
                      default=lambda o: o.model_dump(), sort_keys=False, ensure_ascii=False)


import collections
from dataclasses import dataclass, field
import datetime
from functools import partial
import itertools
import logging.config
import math
from typing import List, Optional, Literal, Dict, Tuple, Union
import logging
import os
import tyro
from ruamel.yaml import YAML
import json

from tqdm import tqdm

from alparc.core.stream import make_streams
from alparc.eval import to_lexicon
from alparc.io import load_phonemes, read_phoneme_corpus, read_syllables_corpus
from alparc.types.base_types import Register, RegisterType
from alparc.types.phoneme import TypePhonemeFeatureLabels
from alparc.types.syllable import LABELS_C, LABELS_V, Syllable
from alparc.types.word import WordType, Word
from alparc.types.lexicon import LexiconType
from alparc.types.stream import StreamType, Stream
from alparc.controls.common import get_oscillation_patterns

from alparc.core.lexicon import make_lexicon_generator, make_lexicons
from alparc.core.word import make_words
from alparc.core.syllable import make_syllables

from alparc.controls.common import *

_OBJECT_DUMP = "_arpac"

def setup_log_dir(results_base_dir: str, name="unknown"):
    results_dir = f"{name}_{datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}"
    results_path = os.path.join(os.path.normpath(results_base_dir), results_dir)
    os.makedirs(results_path, exist_ok=True)
    os.makedirs(os.path.join(results_path, _OBJECT_DUMP), exist_ok=True)
    return results_path


def setup_logging(log_dir: Optional[str] = None, log_console: bool = True, name: str = "unnamed_command") -> Tuple[logging.Logger, str]:
    log_path = setup_log_dir(log_dir, name=name)
    logging.basicConfig(filename=os.path.join(log_path, "debug.log"), 
                        encoding='utf-8', level=logging.DEBUG, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    logger = logging.getLogger(__name__)
    if log_console:
        logger.addHandler(logging.StreamHandler())
    return logger, log_path


def write_stream_summary(streams: Register, save_path: str, logger: logging.Logger):
    yaml = YAML(typ='unsafe', pure=True)
    with open(os.path.join(save_path, "streams.yml"), 'w') as file:
        results = {"streams": {}, "info": {}}
        results["streams"] = [{
                "stream_full": "|".join([syllable.id for syllable in stream]),
                "lexicon": stream.info["lexicon"],
                "lexicon_info": stream.info["lexicon_info"],
                "rhythmicity_indexes": stream.info["rhythmicity_indexes"],
                "stream_tp_mode": stream.info["stream_tp_mode"],
                "n_syllables_per_word": stream.info["n_syllables_per_word"] if "n_syllables_per_word" in stream.info.keys() else None,
                "n_look_back": stream.info["n_look_back"] if "n_look_back" in stream.info.keys() else None,
                "phonotactic_control": stream.info["phonotactic_control"] if "phonotactic_control" in stream.info.keys() else None,
                "syllables_info": stream.info["syllables_info"] if "syllables_info" in stream.info.keys() else None,
        } for stream in streams]
        for stream in streams:
            logger.info(f"- {stream.id}")
        results["info"] = streams.info
        yaml.dump(results, file, encoding="utf-8")

@dataclass
class CommonArgs:
    """"""
    lang: Literal["deu", "eng"] = "deu"
    """The reference language to use (mainly for corpora)"""
    log_dir: Union[str, os.PathLike] = "results"
    """The base directory to safe logs and results to"""
    name: Optional[str] = None
    """Name of the experiment or dataset (to name the subdirectory)"""
    log_console: bool = True
    """Log to console"""
    progress_bars: bool = True
    """Show progress bars in console"""

@dataclass
class SyllableArgs:
    phoneme_pattern: str = "cV"
    """Phoneme pattern to use for syllable generation."""
    unigram_control: bool = True
    """Control for phoneme frequency of use in the syllable compared to the reference language"""
    unigram_alpha: Optional[float] = None
    """Threshold for phoneme frequency of use in the syllable"""
    syllable_control: bool = True
    """Control for syllable frequency of use in the syllable compared to the reference language"""
    syllable_alpha: Optional[float] = None
    """Threshold for syllable frequency of use in the syllable"""
    export_ssml: bool = False
    """Export syllables to SSML format, e.g. for audio generation"""
    consonant_features: List[TypePhonemeFeatureLabels] = field(default_factory=lambda: LABELS_C)
    """Consonant features to use for controls in syllable generation"""
    vowel_features: List[TypePhonemeFeatureLabels] = field(default_factory=lambda: LABELS_V)
    """Vowel features to use for controls in syllable generation"""

@dataclass
class WordArgs:
    n_words: int = 10000
    """Number of words to generate"""
    n_syllables_per_word: Literal[2, 3, 4] = 3
    """Number of syllables per word"""
    bigram_control: bool = True
    """Control for bigram frequency of use in the word compared to the reference language"""
    bigram_alpha: Optional[float] = None
    """Threshold for bigram frequency of use in the word"""
    trigram_control: bool = True
    """Control for trigram frequency of use in the word compared to the reference language"""
    trigram_alpha: Optional[float] = None
    """Threshold for trigram frequency of use in the word"""
    positional_control: bool = True
    """Control for positional frequency of use of a phoneme in the word compared to the reference language"""
    positional_control_position: Optional[int] = None
    """Position of the phoneme in the word (0 = first, 1 = second, ...). If None, all positions are controlled"""
    position_alpha: int = 0
    """Threshold for positional frequency of use of a phoneme in the word"""
    phonotactic_control: bool = True
    """Control for phonotactic feature repetition of the phonemes in the word"""
    n_look_back: int = 2
    """Number of phonemes to look back for phonotactic control"""
    max_tries: int = 100000
    """Maximum number of tries to generate the word register with the given constraints"""

@dataclass
class LexiconArgs:
    n_lexicons: int = 2
    """Number of lexicons to generate"""
    n_words_per_lexicon: Literal[3, 4, 5] = 4
    """Number of words per lexicon"""
    unique_words: bool = False
    """Check uniqueness of words across all lexicons"""
    binary_feature_control: bool = True
    """Control for binary feature repetition between words in the lexicon. 
    See 'lag_of_interest', 'max_overlap', and 'max_word_matrix'."""
    lag_of_interest: int = 1
    """Binary feature frequency in words"""
    max_overlap: int = 1
    """Maximum number of overlapping features between words in the lexicon"""
    max_word_matrix: int = 200
    """Maximum number of words to use to create pairwise feature overlaps (Will be sub-sampled if necessary)"""
    control_features: List[TypePhonemeFeatureLabels] = field(default_factory=lambda: LABELS_C + LABELS_V)
    """If controlled, which binary features to include in binary feature control"""

@dataclass
class StreamArgs:
    repetitions: int = 15
    """Number of repetitions of the lexicon contents to create a full stream"""
    max_rhythmicity: Optional[float] = None
    """Threshold for maximum rhythmicity index of features in the stream. If None, rhythmicity control is still applied, but no threshold"""
    n_streams_per_lexicon: int = 2
    """Number of streams to generate per lexicon"""
    max_tries_randomize: int = 10
    """Maximum number of tries to randomize the stream (only if max_rhythmicity is used)"""
    tp_modes: List[Literal["random", "word_structured", "position_controlled"]] = field(default_factory=lambda: ["random", "word_structured", "position_controlled"])
    """Rules to use for the syllable randomization. If None, all patterns are used"""
    require_all_tp_modes: bool = True
    """If True, all tp_modes are required to return a valid stream for a given lexicon, otherwise the stream will be dropped"""

@dataclass
class Generate:
    """Generate a dataset of streams from a phenome database and language-specific phoneme, syllable and n-gram corpora"""
    common: CommonArgs
    syllable: SyllableArgs
    word: WordArgs
    lexicon: LexiconArgs
    stream: StreamArgs

def generate_stream_dataset(args: Generate) -> RegisterType:
    logger, log_dir = setup_logging(args.common.log_dir, args.common.log_console, name=args.common.name or "generate_streams")
    
    with open(os.path.join(log_dir, "config.yml"), "w") as file:
        yaml = YAML(typ='unsafe', pure=True)
        yaml.dump(vars(args), file, encoding="utf-8")

    phonemes = load_phonemes(lang=(args.common.lang if args.syllable.unigram_control else None))
    if args.syllable.unigram_alpha is not None:
        phonemes = phonemes.filter(lambda unigram: unigram.info["p_unif"] > args.syllable.unigram_alpha)
    logger.info(f"Generate Phonemes: {phonemes}")

    syllables = make_syllables(
        phonemes=phonemes, 
        phoneme_pattern=args.syllable.phoneme_pattern,
        syllable_control=args.syllable.syllable_control,
        syllable_alpha=args.syllable.syllable_alpha,
        lang=args.common.lang,
        consonant_features=args.syllable.consonant_features,
        vowel_features=args.syllable.vowel_features,
    )
    logger.info(f"Generate Syllables: {syllables}")

    syllables.save(os.path.join(log_dir, _OBJECT_DUMP, "syllables.json"))

    if args.syllable.export_ssml:
        from alparc.io import export_speech_synthesizer
        export_speech_synthesizer(syllables, syllables_dir=os.path.join(log_dir, "ssml"))

    logger.info(f"Generate Pseudo-Words: ...")
    pseudo_words = make_words(
        syllables=syllables,
        num_syllables=args.word.n_syllables_per_word,
        bigram_control=args.word.bigram_control,
        bigram_alpha=args.word.bigram_alpha,
        trigram_control=args.word.trigram_control,
        trigram_alpha=args.word.trigram_alpha,
        positional_control=args.word.positional_control,
        positional_control_position=args.word.positional_control_position,
        position_alpha=args.word.position_alpha,
        phonotactic_control=args.word.phonotactic_control,
        n_look_back=args.word.n_look_back,
        n_words=args.word.n_words,
        max_tries=args.word.max_tries,
        progress_bar=args.common.progress_bars,
        lang=args.common.lang,
    )
    logger.info(f"Pseudo-Words: {pseudo_words}")

    pseudo_words.save(os.path.join(log_dir, _OBJECT_DUMP, "pseudo_words.json"))

    logger.info(f"Generate Lexicons: ...")
    lexicons = make_lexicons(
        pseudo_words, 
        n_lexicons=args.lexicon.n_lexicons, 
        n_words=args.lexicon.n_words_per_lexicon,
        max_overlap=args.lexicon.max_overlap,
        lag_of_interest=args.lexicon.lag_of_interest,
        max_word_matrix=args.lexicon.max_word_matrix,
        unique_words=args.lexicon.unique_words,
        control_features=args.lexicon.control_features,
        progress_bar=args.common.progress_bars,
        binary_feature_control=args.lexicon.binary_feature_control,
    )
    logger.info(f"Lexicons: {[str(l) for l in lexicons]}")

    for i, lexicon in enumerate(lexicons):
        lexicon.save(os.path.join(log_dir, _OBJECT_DUMP, f"lexicon_{i}.json"))

    logger.info(f"Generate Streams: ...")
    streams = Register()
    for _ in tqdm(range(args.stream.n_streams_per_lexicon)):
        for stream in make_streams(
            lexicons,
            max_rhythmicity=args.stream.max_rhythmicity,
            stream_length=args.stream.repetitions,
            max_tries_randomize=args.stream.max_tries_randomize,
            tp_modes=args.stream.tp_modes,
            require_all_tp_modes=args.stream.require_all_tp_modes
        ):
            streams.append(stream)
    
    logger.info(f"Streams: ")
    streams.save(os.path.join(log_dir, _OBJECT_DUMP, f"streams.json"))
    write_stream_summary(streams, save_path=log_dir, logger=logger)


def write_lexicon_summary(lexicon: Register, save_path: str, logger: logging.Logger):
    with open(os.path.join(save_path, "lexicons.yml"), 'w') as file:
        results = {}
        results["lexicon"] = "|".join(word.id for word in lexicon)
        results["info"] = lexicon.info
        yaml = YAML(typ='unsafe', pure=True)
        yaml.dump(results, file, encoding="utf-8")


@dataclass
class Diagnose:
    """Diagnose a lexicon by checking its phonotactic, acoustic and rhythmic properties"""
    lexicons: str
    """Lexicon string(s) consisting of words and syllables. Multiple lexicons should be separated by ' '.
    Syllables should be separated by '|' and words by '||'. Example: pi|ɾu|ta||ba|ɡo|li||to|ku|da||ɡu|haɪ|bo"""
    common: CommonArgs
    stream: StreamArgs
    export_ssml: bool = True
    """Export syllables to SSML format, e.g. for audio generation"""
    split_registers: bool = False
    """Derive phoneme and syllable registers from the lexicon"""
    generate_streams: bool = True
    """Generate streams from the parsed lexicons"""
    phoneme_pattern: str = "cv"
    """Phoneme pattern to assume for syllable parsing"""

def evaluate_lexicons(args: Diagnose):
    logger, log_dir = setup_logging(args.common.log_dir, args.common.log_console, name=args.common.name or "evaluate_lexicon")

    yaml = YAML(typ='unsafe', pure=True)
    with open(os.path.join(log_dir, "config.yml"), "w") as file:
        yaml.dump(vars(args), file, encoding="utf-8")

    lexicons = [[w.split("|") for w in l.split("||")] for l in args.lexicons.split(" ")]
    lexicons = list(map(partial(to_lexicon, syllable_type=args.phoneme_pattern), lexicons))
    save_path = os.path.join(log_dir, _OBJECT_DUMP, "lexicon.json")
    for lexicon in lexicons:
            logger.info(f"Read Lexicon: {lexicon}")
            lexicon.save(save_path)
            logger.info(f"Lexicon object saved to file: {save_path}")
            save_path = os.path.join(log_dir, "lexicon.yml")
            with open(save_path, 'w') as file:
                summary = {}
                summary["lexicon"] = "|".join(word.id for word in lexicon)
                summary["info"] = lexicon.info
                yaml.dump(summary, file, encoding="utf-8")
            logger.info(f"Lexicon summary saved to file: {save_path}")

    if args.generate_streams:
        logger.info(f"Generate Streams: ...")
        streams = Register()
        for _ in tqdm(range(args.stream.n_streams_per_lexicon)):
            for stream in make_streams(
                lexicons,
                max_rhythmicity=args.stream.max_rhythmicity,
                stream_length=args.stream.repetitions,
                max_tries_randomize=args.stream.max_tries_randomize,
                tp_modes=args.stream.tp_modes,
                require_all_tp_modes=args.stream.require_all_tp_modes
            ):
                streams.append(stream)
        
        logger.info(f"Streams: ")
        streams.save(os.path.join(log_dir, _OBJECT_DUMP, f"streams.json"))
        write_stream_summary(streams, save_path=log_dir, logger=logger)

    if args.split_registers:
        for lexicon in lexicons:
            syllables = lexicon.flatten()
            syllables.save(os.path.join(log_dir, _OBJECT_DUMP, "syllables.json"))
            logger.info(f"Syllables object saved to file: {os.path.join(log_dir, _OBJECT_DUMP, 'syllables.json')}")

            syllables_with_corpus_stats = syllables.intersection(read_syllables_corpus(lang=args.common.lang))
            syllables_with_corpus_stats.save(os.path.join(log_dir, _OBJECT_DUMP, "syllables_with_corpus_stats.json"))
            logger.info(f"Syllables object with corpus stats saved to file: {os.path.join(log_dir, _OBJECT_DUMP, 'syllables_with_corpus_stats.json')}")

            if args.export_ssml:
                from alparc.io import export_speech_synthesizer
                export_speech_synthesizer(syllables, syllables_dir=os.path.join(log_dir, "ssml"))

            phonemes = syllables.flatten()
            phonemes.save(os.path.join(log_dir, _OBJECT_DUMP, "phonemes.json"))
            logger.info(f"Phonemes object saved to file: {os.path.join(log_dir, _OBJECT_DUMP, 'phonemes.json')}")

            if args.common.lang == "deu":
                corpus_phonemes = read_phoneme_corpus(lang=args.common.lang)
                phonemes_with_german_corpus_stats = phonemes.intersection(corpus_phonemes)
                phonemes_with_german_corpus_stats.save(os.path.join(log_dir, _OBJECT_DUMP, "phonemes_with_german_corpus_stats.json"))
                logger.info(f"Phonemes object with corpus stats saved to file: {os.path.join(log_dir, _OBJECT_DUMP, 'phonemes_with_german_corpus_stats.json')}")


@dataclass
class EvaluateStream:
    stream: str
    """Stream string consisting of syllables, separated by '|'. 
    Example: pi|ɾu|ta|ba|ɡo|li|to|li|to|ku|ɾu|ta|ba|ɡo|li|to|ku|da|ɡu|ki|bo"""


def cli():
    args = tyro.cli(Union[Generate, Diagnose], prog="alparc", description="The ALPARC Toolbox: Artificial Languages with Phonological and Acoustic Rhythmicity Control")
    if isinstance(args, Generate):
        generate_stream_dataset(args)
    if isinstance(args, Diagnose):
        evaluate_lexicons(args)



import csv
import json
import logging
import os
import pathlib
from importlib import resources as importlib_resources
from os import PathLike
from typing import Iterable, Dict, Union, List, Type, Optional, Literal
from functools import partial
from copy import copy

import pandas as pd
import numpy as np
from scipy import stats

from alparc.phonecodes import phonecodes
from alparc.types.base_types import Register, RegisterType

from alparc.types.phoneme import PHONEME_FEATURE_LABELS, Phoneme
from alparc.types.syllable import Syllable
from alparc.types.word import Word
from alparc.types.lexicon import Lexicon, LexiconType
from alparc.types.stream import Stream

logger = logging.getLogger(__name__)


def get_data_path(fname):
    return importlib_resources.files("alparc") / "data" / fname


BINARY_FEATURES_DEFAULT_PATH = get_data_path("phonemes.csv")
PHONEMES_DEFAULT_PATH = get_data_path("phonemes.json")

CORPUS_DEFAULT_PATH_DEU_SPECIAL = get_data_path("german")
SYLLABLES_DEFAULT_PATH_DEU_SPECIAL = CORPUS_DEFAULT_PATH_DEU_SPECIAL / 'syllables.csv'
IPA_BIGRAMS_DEFAULT_PATH = CORPUS_DEFAULT_PATH_DEU_SPECIAL / 'bigrams.csv'
IPA_TRIGRAMS_DEFAULT_PATH = CORPUS_DEFAULT_PATH_DEU_SPECIAL / 'trigrams.csv'
IPA_SEG_DEFAULT_PATH = CORPUS_DEFAULT_PATH_DEU_SPECIAL / 'unigrams.csv'

CORPUS_DEFAULT_PATH_CELEX = get_data_path("english")
SYLLABLES_DEFAULT_PATH_ENG = CORPUS_DEFAULT_PATH_CELEX / "syllables.csv"
#SYLLABLES_DEFAULT_PATH_DEU = CORPUS_DEFAULT_PATH_CELEX / "GERMAN" / "EFS" / "EFS.CD"
# SYLLABLES_DEFAULT_PATH_NLD = CORPUS_DEFAULT_PATH_CELEX / "DUTCH" / "EFS" / "EFS.CD"

RESULTS_DEFAULT_PATH = pathlib.Path("arc_results")
SSML_RESULTS_DEFAULT_PATH = RESULTS_DEFAULT_PATH / "syllables"


def export_speech_synthesizer(syllables: Iterable[Syllable],
                              syllables_dir: Union[str, PathLike] = SSML_RESULTS_DEFAULT_PATH):
    logger.info("SAVE EACH SYLLABLE TO A TEXT FILE FOR THE SPEECH SYNTHESIZER")
    os.makedirs(syllables_dir, exist_ok=True)
    c = [s.id[0] for s in syllables]
    v = [s.id[1] for s in syllables]
    c = ' '.join(c).replace('ʃ', 'sch').replace('ɡ', 'g').replace('ç', 'ch').replace('ʒ', 'dsch').split()
    v = ' '.join(v).replace('ɛ', 'ä').replace('ø', 'ö').replace('y', 'ü').split()
    t = [co + vo for co, vo in zip(c, v)]
    for syllable, text in zip(syllables, t):
        synth_string = '<phoneme alphabet="ipa" ph=' + '"' + syllable.id + '"' + '>' + text + '</phoneme>'
        with open(os.path.join(syllables_dir, f'{str(syllable.id[0:2])}.txt'), 'w') as f:
            f.write(synth_string + "\n")
            csv.writer(f)

    print("Done")


def read_phoneme_corpus(
        lang: Literal["deu", "eng"] = "eng",
) -> Register[str, Phoneme]:
    """
    Read order of phonemes, i.e. phonemes from a corpus together with the positions at which they
    appear in a bag of words.

    :param ipa_seg_path:
    :return:
    """
    if not (lang == "deu"):
        logger.warning("Only german phoneme corpus available")
        return read_default_phonemes()

    logger.info("READ ORDER OF PHONEMES IN WORDS")

    # TODO: make language specific
    ipa_seg_path: Union[os.PathLike, str] = IPA_SEG_DEFAULT_PATH

    with open(ipa_seg_path, "r", encoding='utf-8') as csv_file:
        fdata = list(csv.reader(csv_file))[1:]

    phonemes = {}
    for phon, position_in_word in fdata:
        phon = phon.replace('"', '').replace("g", "ɡ")
        position_in_word = int(position_in_word)
        if phon in phonemes:
            phonemes[phon].info["order"].append(position_in_word)
        else:
            phonemes[phon] = Phoneme(id=phon, info={"order": [position_in_word]})

    for phon in phonemes:
        positions = max(phonemes[phon].info["order"])
        phonemes[phon].info["word_position_prob"] = {}
        for position in range(positions):
            phoneme_pos_prob = phonemes[phon].info["order"].count(position + 1) / len(phonemes[phon].info["order"])
            phonemes[phon].info["word_position_prob"][position] = phoneme_pos_prob
        del phonemes[phon].info["order"]

    return Register(phonemes)


def syll_to_ipa(syll, language="deu", from_format="xsampa"):
    if from_format == "xsampa":
        return phonecodes.xsampa2ipa(syll, language)
    elif from_format == "ipa":
        return syll
    else:
        raise ValueError(f"Unknown format {from_format}")


def read_syllables_corpus(
        lang: str = "deu",
) -> Register[str, Syllable]:
    logger.info("READ SYLLABLES, FREQUENCIES AND PROBABILITIES FROM CORPUS AND CONVERT SYLLABLES TO IPA")

    if lang == "deu":
        syllables_corpus_path: Union[os.PathLike, str] = SYLLABLES_DEFAULT_PATH_DEU_SPECIAL
    elif lang == "eng":
        syllables_corpus_path: Union[os.PathLike, str] = SYLLABLES_DEFAULT_PATH_ENG
    else:
        raise ValueError(f"Language {lang} not supported.")
    
    with open(syllables_corpus_path, "r", encoding='utf-8') as csv_file:
        data = list(csv.reader(csv_file))[1:]

    syllables_dict: Dict[str, Syllable] = {}

    for syll_ipa, freq, prob in data:
        info = {"freq": int(freq), "prob": float(prob)}
        if syll_ipa not in syllables_dict or syllables_dict[syll_ipa].info != info:
            syllables_dict[syll_ipa] = Syllable(
                id=syll_ipa, phonemes=[], info=info, binary_features=[], phonotactic_features=[])
        else:
            logger.info(
                f"Syllable '{syll_ipa}' with conflicting stats {info} != {syllables_dict[syll_ipa].info}."
            )

    return Register(syllables_dict)


def read_bigrams(
    ipa_bigrams_path: str = IPA_BIGRAMS_DEFAULT_PATH,
) -> Register[str, Syllable]:
    logger.info("READ BIGRAMS")

    with open(ipa_bigrams_path, "r", encoding='utf-8') as csv_file:
        fdata = list(csv.reader(csv_file))[1:]

    freqs = [int(data[1]) for data in fdata]
    p_vals_uniform = stats.uniform.sf(abs(stats.zscore(np.log(freqs))))

    bigrams_dict: Dict[str, Syllable] = {}

    for (bigram, freq), p_unif in zip(fdata, p_vals_uniform):
        bigram = bigram.replace('_', '').replace("g", "ɡ")
        info = {"freq": int(freq), "p_unif": p_unif}

        if bigram not in bigrams_dict or bigrams_dict[bigram].info == info:
            # a bigram is not necessarily a syllable but in our type system they are equivalent
            bigrams_dict[bigram] = Syllable(id=bigram, phonemes=[], info=info)
        else:
            logger.info(
                f"Bigram '{bigram}' with conflicting stats {info} != {bigrams_dict[bigram].info}."
            )
            # del bigrams_dict[bigram]

    return Register(bigrams_dict)


def read_trigrams(
        ipa_trigrams_path: str = IPA_TRIGRAMS_DEFAULT_PATH,
) -> Register[str, Syllable]:
    logger.info("READ TRIGRAMS")
    
    with open(ipa_trigrams_path, "r", encoding='utf-8') as csv_file:
        fdata = list(csv.reader(csv_file))[1:]

    freqs = [int(data[1]) for data in fdata]
    p_vals_uniform = stats.uniform.sf(abs(stats.zscore(np.log(freqs))))

    trigrams = Register()
    for (trigram, freq), p_unif in zip(fdata[1:], p_vals_uniform):
        trigram = trigram.replace('_', '').replace("g", "ɡ")
        info = {"freq": int(freq), "p_unif": p_unif}

        if trigram not in trigrams or trigrams[trigram].info == info:
            trigrams[trigram] = Syllable(id=trigram, phonemes=[], info=info)
        else:
            logger.info(
                f"Trigram '{trigram}' with conflicting stats {info} != {trigrams[trigram].info}."
            )

    return trigrams


def read_default_phonemes() -> Register:
    logger.info("READ MATRIX OF BINARY FEATURES FOR ALL IPA PHONEMES")

    with open(BINARY_FEATURES_DEFAULT_PATH, "r", encoding='utf-8') as csv_file:
        fdata = list(csv.reader(csv_file))

    phons = [row[0] for row in fdata[1:]]
    feats = [row[1:] for row in fdata[1:]]
    phoneme_feature_labels = fdata[0][1:]

    assert phoneme_feature_labels == PHONEME_FEATURE_LABELS

    phonemes_dict = {}
    for phon, features in zip(phons, feats):
        if phon not in phonemes_dict or features == phonemes_dict[phon].info["features"]:
            phonemes_dict[phon] = Phoneme(id=phon, info={"features": features})
        else:
            logger.info(
                f"Phoneme '{phon}' with conflicting "
                f"feature entries {features} != {phonemes_dict[phon].info['features']}.")
            # del phonemes_dict[phon]

    return Register(phonemes_dict, _info={"phoneme_feature_labels": phoneme_feature_labels})


def check_german(words: List[Word]):
    # TODO
    # SAVE WORDS IN ONE CSV FILE
    with open(os.path.join(RESULTS_DEFAULT_PATH, 'words.csv'), 'w', encoding='utf-8') as f:
        writer = csv.writer(f)
        for word in words:
            writer.writerows([[word.id, 0]])

    # TO ENSURE THAT THE TRIPLETS CANNOT BE MISTAKEN FOR GERMAN WORDS,
    # WE INSTRUCTED A NATIVE GERMAN SPEAKER TO MARK EACH TRIPLET AS...
    #     '1' IF IT CORRESPONDS EXACTLY TO A GERMAN WORD
    #     '2' IF IT COULD BE MISTAKEN FOR A GERMAN WORD-GROUP WHEN PRONOUNCED ALOUD
    #     '3' IF THE PRONUNCIATION OF THE FIRST TWO SYLLABLES IS A WORD CANDIDATE,
    #         i.e. the syllable pair could be mistaken for a German word, or
    #         it evokes a strong prediction for a certain real German word
    #     '4' IF IT DOES NOT SOUND GERMAN AT ALL
    #         (that is, if the phoneme combination is illegal in German morphology
    #         [do not flag if rule exceptions exist])
    #     '0' OTHERWISE (that is, the item is good)

    logger.info("LOAD WORDS FROM CSV FILE AND SELECT THOSE THAT CANNOT BE MISTAKEN FOR GERMAN WORDS")
    with open(os.path.join(RESULTS_DEFAULT_PATH, "words.csv"), 'r', encoding='utf-8') as f:
        fdata = list(csv.reader(f, delimiter='\t'))
    rows = [row[0].split(",") for row in fdata]
    words = [row[0] for row in rows if row[1] == "0"]

    return words


def arc_register_from_json(path: Union[str, PathLike], arc_type: Type) -> RegisterType:
    """
    Load an arc register from a json file.
    """
    with open(path, "r", encoding='utf-8') as file:
        d = json.load(file)

    # we have to process the "_info" field separately because it's not a valid ARC type
    register = Register({k: arc_type(**v) for k, v in d.items() if k != "_info"})
    register.info = d["_info"]

    return register


def load_phonemes(lang: Optional[Literal["deu", "eng"]] = "deu") -> RegisterType:
    """Load phoneme corpus with phonological features.

    Args:
        language_control (bool, optional): Filter out phonemes that are not common in the phoneme corpus (german). Defaults to True.

    Returns:
        RegisterType: _description_
    """
    phonemes = read_default_phonemes()
    
    if lang is not None:
        phonemes = phonemes.intersection(read_phoneme_corpus(lang=lang))

    return phonemes


def load_syllables(path_to_json: Union[str, PathLike]):
    return arc_register_from_json(path_to_json, Syllable)


def load_words(path_to_json: Union[str, PathLike]):
    return arc_register_from_json(path_to_json, Word)

def load_lexicons(path_to_json: Union[str, PathLike]):
    return arc_register_from_json(path_to_json, LexiconType)

def load_streams(path_to_json: Union[str, PathLike]):
    register = arc_register_from_json(path_to_json, Stream)
    #for stream in register:
    #    if "lexicon" in stream.info.keys():
    #        data = stream.info["lexicon"]
    #        print(data)
    #        lexicon = Lexicon({k: Word(**v) for k, v in data.items() if k != "info"})
    #        lexicon.info = data["info"]
    #        stream.info["lexicon"] = lexicon
    return register



import collections
import itertools
import logging.config
import math
from typing import List, Optional, Literal, Dict, Union
import logging
import os

from alparc.io import load_phonemes
from alparc.types.base_types import Register, RegisterType
from alparc.types.syllable import Syllable
from alparc.types.word import WordType, Word
from alparc.types.lexicon import LexiconType
from alparc.types.stream import StreamType, Stream
from alparc.controls.common import get_oscillation_patterns

from alparc.core.lexicon import make_lexicon_generator, make_lexicons
from alparc.core.word import make_words
from alparc.core.syllable import make_syllables

from alparc.controls.common import *


logger = logging.getLogger(__name__)


def transitional_p_matrix(v):
    # TODO: make (faster and) easier to read
    n = 1 + max(v)
    M = [[0] * n for _ in range(n)]
    for (i, j) in zip(v, v[1:]):
        M[i][j] += 1
    for r in M:
        s = sum(r)
        if s > 0:
            r[:] = [i/s for i in r]
    return M

def shuffled_struct_stream(n_words=4, n_sylls_per_word=3, n_repetitions=15):
    # TODO: make faster and easier to read
    n_sylls_total = n_sylls_per_word * n_words  # number of syllables in a lexicon
    n_iters = n_words * n_repetitions  # number of repetitions in a trial
    syll_id = np.arange(n_sylls_total).reshape((n_words, int(n_sylls_total / n_words)))
    while True:
        stream = np.repeat(syll_id, n_iters, axis=0)
        np.random.shuffle(stream)
        r = list(stream)
        for n in range(1, len(r)-1):
            if r[n][0] == r[n-1][0]:
                k = r[n:]
                l = [i for i, el in enumerate(k) if el[0] != r[n][0]]
                if not l:
                    continue
                else:
                    l = l[0]
                r[n] = r[n+l]
                r[n+l] = r[n-1]
        x = [i[0] for i in r]
        check = [sum(1 for _ in group) for _, group in itertools.groupby(x)]
        if all(i == 1 for i in check):
            break
    v = [j for k in r for j in k]
    v.append(v[0])
    M = np.array(transitional_p_matrix(v))
    v.pop()
    return v, M

def shuffled_uniform_stream(n_words=4, n_sylls_per_word=3, n_repetitions=15):
    # TODO: make faster and easier to read
    n_sylls_total = n_sylls_per_word * n_words  # number of syllables in a lexicon
    n_iters = n_words * n_repetitions  # number of repetitions in a trial:
    syll_id = np.arange(n_sylls_total)
    while True:
        stream = np.repeat(syll_id, n_iters)
        np.random.shuffle(stream)
        v = list(stream)
        for n in range(1, len(v)-1):
            if v[n] == v[n-1]:
                v[n] = v[n+1]
                v[n+1] = v[n-1]
        check = [sum(1 for _ in group) for _, group in itertools.groupby(v)]
        if all(i == 1 for i in check):
            break
    v.append(v[0])
    M = np.array(transitional_p_matrix(v))
    v.pop()
    return v, M

def pseudo_rand_tp_uniform(n_words=4, n_sylls_per_word=3, n_repetitions=15):
    # TODO: make faster and easier to read
    n_sylls_total = n_sylls_per_word * n_words  # number of syllables in a lexicon
    n_iters = n_words * n_repetitions  # number of repetitions in a trial
    n_loop = 1000  # ???
    while True:
        p1 = np.repeat(range(n_sylls_total), math.ceil(n_iters / int(n_sylls_total / n_words))).tolist()
        p2 = np.repeat(range(n_sylls_total), math.ceil(n_iters / int(n_sylls_total / n_words))).tolist()
        p3 = np.repeat(range(n_sylls_total), math.ceil(n_iters / int(n_sylls_total / n_words))).tolist()
        M = np.zeros((n_sylls_total, n_sylls_total))
        mask = (np.ones((n_sylls_total, n_sylls_total)) - np.diag(np.ones(n_sylls_total))).astype(bool)
        v = []
        for _ in range(n_iters):
            if not v:
                p = random.sample(range(n_sylls_total), n_sylls_total)
                m = np.array(transitional_p_matrix(p))
            else:
                i_loop = 0
                while i_loop < n_loop:
                    try:
                        s = v[-1]
                        p = []
                        m = np.zeros((n_sylls_total, n_sylls_total))
                        N = np.copy(M)
                        for idx in range(n_sylls_total):
                            t = list(np.where(N[s] == N[s].min())[0])
                            t.remove(s)
                            if idx in list(range(n_sylls_total)[0::3]):
                                t = [i for i in p1 if i in t and i not in p]
                                if not t and i_loop > n_loop/100:
                                    t = list(np.where(N[s] == np.sort(N[s])[1])[0])
                                    t = [i for i in p1 if i in t and i not in p]
                                if not t and i_loop > n_loop/10:
                                    t = [i for i in p1 if i != s]                        
                                    t = [i for i in p1 if i in t and i not in p]
                                x = np.array(list(map(list(t).count, set(t))))
                                n = [list(set(t))[i] for i in list(np.where(x == x.max())[0])]
                                p += random.sample(n, 1)
                            elif idx in list(range(n_sylls_total)[1::3]):
                                t = [i for i in p2 if i in t and i not in p]
                                if not t and i_loop > n_loop/100:
                                    t = list(np.where(N[s] == np.sort(N[s])[1])[0])
                                    t = [i for i in p2 if i in t and i not in p]
                                if not t and i_loop > n_loop/10:
                                    t = [i for i in p2 if i != s]
                                    t = [i for i in p2 if i in t and i not in p]
                                x = np.array(list(map(list(t).count, set(t))))
                                n = [list(set(t))[i] for i in list(np.where(x == x.max())[0])]
                                p += random.sample(n, 1)
                            elif idx in list(range(n_sylls_total)[2::3]):
                                t = [i for i in p3 if i in t and i not in p]
                                if not t and i_loop > n_loop/100:
                                    t = list(np.where(N[s] == np.sort(N[s])[1])[0])
                                    t = [i for i in p3 if i in t and i not in p]
                                if not t and i_loop > n_loop/10:
                                    t = [i for i in p3 if i != s]
                                    t = [i for i in p3 if i in t and i not in p]
                                x = np.array(list(map(list(t).count, set(t))))
                                n = [list(set(t))[i] for i in list(np.where(x == x.max())[0])]
                                p += random.sample(n, 1)    
                            m[s][p[idx]] = m[s][p[idx]] + 1
                            s = p[-1]
                            N += m
                        break
                    except:
                        i_loop += 1
            if len(p) == n_sylls_total:
                M += m
                v += p
            else:
                break
            if all(i in p1 
                   for i in p[0::3]) and all(i in p2 
                                             for i in p[1::3]) and all(i in p3 
                                                                       for i in p[2::3]):
                for i in range(n_words):
                    p1.remove(p[0::3][i])
                    p2.remove(p[1::3][i])
                    p3.remove(p[2::3][i])
            else:
                break
        if all(np.diagonal(M) == np.zeros(n_sylls_total)) and len(v) == n_sylls_total * n_iters:
            if M[mask].min() == math.floor(n_iters * n_sylls_total / len(M[mask])):
                if M[mask].max() == math.ceil(n_iters * n_sylls_total / len(M[mask])):
                    M = np.array(transitional_p_matrix(v))
                    break
    return v, M

# subfunctin for pseudo_rand_tp_random_position_controlled()
def pseudo_walk_tp_uniform(P, T, v, S, N, n_words=4, n_sylls_per_word=3):
    # TODO: make faster and easier to read
    t = []
    for iTrip in range(n_words):
        for iPoss in range(n_sylls_per_word):
            if not v and not t:
                if N:
                    x = [N]
                    N = []
                else:
                    x = random.sample(P[iPoss], 1)
                t += x
            else:
                iLoop = 0
                while iLoop < 1000:
                    x = random.sample([i for i in P[iPoss] if i not in t], 1)[0]
                    if not t:
                        s = v[-1]
                    else:
                        s = t[-1]
                    if [s, x] not in S and [s, x] in T[iPoss-1]:
                        S.append([s, x])
                        t.append(x)
                        break
                    else:
                        iLoop += 1
    if iLoop < 1000:
        v += t
    return v, S, t

def pseudo_rand_tp_uniform_position_controlled(n_words=4, n_sylls_per_word=3, n_repetitions=15):
    # TODO: make faster and easier to read
    n_sylls_total = n_sylls_per_word * n_words  # number of syllables in a lexicon
    n_iters = n_words * n_repetitions  # number of repetitions in a trial
    P = [list(range(i, n_sylls_total, n_sylls_per_word)) for i in range(n_sylls_per_word)]
    I = list(range(n_sylls_per_word))
    B = I[:]
    B.append(B.pop(0))
    T = [[list(i) for i in list(itertools.product(*[P[j], P[k]]))] for j, k in zip(I, B)]
    while True:
        V = []
        while len(V) < n_sylls_total * n_iters:
            v = []
            t = []
            if not V:
                N = []
                S = []
            else:
                N = [i for i in T[n_sylls_per_word - 1] if i not in S][0][1]
                S = []
            while len(v) < len(T) * len(T[0]):
                v, S, t = pseudo_walk_tp_uniform(P, T, v, S, N, n_words, n_sylls_per_word)
                if len(t) < n_sylls_total:
                    v = []
                    S = []
                    t = []
            V += v
        V.append(V[0])
        M = np.array(transitional_p_matrix(V))
        V.pop()
        if all(set(i) <= set([1/n_words, 0]) for i in M):
            break
    return V, M

def pseudo_rand_tp_struct(n_words=4, n_sylls_per_word=3, n_repetitions=15):
    # TODO: make faster and easier to read
    n_sylls_total = n_sylls_per_word * n_words  # number of syllables in a lexicon
    n_iters = n_words * n_repetitions  # number of repetitions in a trial
    while True:
        P = list(itertools.permutations(np.arange(n_words)))
        S = [[*(i for _ in range(math.ceil(n_iters / len(P))))] for i in P]
        M = np.zeros((n_words, n_words))
        v = []
        c = 2
        mask = (np.ones((n_words, n_words)) - np.diag(np.ones(n_words))).astype(bool)
        for idx in range(n_iters):
            R = [S[i] for i in range(len(S)) if len(S[i]) == max(map(len, S))]
            if not v:
                p = list(list(random.sample(R, 1)[0])[0])
                m = np.array(transitional_p_matrix(p))
            else:
                s = v[-1]
                n = list(np.where(M[s] == M[s].min())[0])
                n.remove(s)
                if not n:
                    n = list(np.where(M[s] == np.sort(M[s])[1])[0])
                T = [j for j in P if any(k for k in [(s, i) for i in n]) in bytes(j) and j[0] != s]
                t = [i[0] for i in [[i for i in T if i in j] for j in R] if i]
                if not t:
                    t = T
                while True:
                    if collections.Counter(M[mask])[c - 1] >= len(M[mask]) - n_words:
                        c += 1
                    if t:
                        p = list(random.sample(t, 1)[0])
                        m = np.array(transitional_p_matrix(p))
                        m[s][p[0]] = m[s][p[0]] + 1
                        if c not in M + m:
                            break
                        else:
                            t.remove(tuple(p))
                    else:
                        c += 1
                        T = [i for i in P if any(j for j in [(s, k) for k in n]) in bytes(i) and i[0] != s]
                        t = [i[0] for i in [[i for i in T if i in j] for j in R] if i]
                        if not t:
                            t = T
            M += m
            v += p
            r = P.index(tuple(p))
            if tuple(p) in S[r]:
                S[r].remove(tuple(p))
            else:
                break
        if all(np.diagonal(M) == np.zeros(n_words)) and len(v) == n_words * n_iters:
            if M[mask].min() == math.floor(n_iters * n_words / len(M[mask])) - 1:
                if M[mask].max() == math.ceil(n_iters * n_words / len(M[mask])):
                    v.append(v[0])
                    M = np.array(transitional_p_matrix(v))
                    v.pop()
                    break
    return v, M


def sample_syllable_randomization(
        lexicon: LexiconType,
        n_repetitions: int = 4,
        max_tries=10,
        tp_mode: Literal["word_structured", "position_controlled", "random"] = "word_structured"):
    if tp_mode == "word_structured":
        elements = [word for word in lexicon]
        rand_func = pseudo_rand_tp_struct
    elif tp_mode == "position_controlled":
        elements = [syllable for word in lexicon for syllable in word]
        rand_func = pseudo_rand_tp_uniform_position_controlled
    elif tp_mode == "random":
        elements = [syllable for word in lexicon for syllable in word]
        rand_func = pseudo_rand_tp_uniform
    else:
        raise ValueError(f"tp_mode '{tp_mode}' unknown.")

    randomized_indexes_list = []
    n_syllables = len(lexicon[0].syllables)
    n_words = len(lexicon)

    for _ in range(max_tries):
        randomized_indexes, _ = rand_func(n_words=n_words, n_sylls_per_word=n_syllables, n_repetitions=n_repetitions)
        if randomized_indexes not in randomized_indexes_list:
            randomized_indexes_list.append(randomized_indexes)
            if tp_mode == "word_structured":
                yield [syllable for index in randomized_indexes for syllable in elements[index]]
            else:
                yield [elements[index] for index in randomized_indexes]


def compute_rhythmicity_index_sylls_stream(stream, patterns):
    count_patterns = []
    patterns = [tuple(pat) for pat in patterns]
    for feature_stream in list(zip(*[syllable.info["binary_features"] for syllable in stream])):
        c = 0
        for iSyll in range(len(feature_stream) - max(len(i) for i in patterns)):
            if any(i == feature_stream[iSyll: iSyll + len(i)] for i in patterns):
                c += 1
        count_patterns.append(c / (len(feature_stream) - max(len(i) for i in patterns)))

    return count_patterns


def make_stream_from_lexicon(lexicon: Register[str, Word],
                             max_rhythmicity: Optional[float] = None, max_tries_randomize=10, n_repetitions: int = 4,
                             tp_mode: Literal["word_structured", "position_controlled", "random"] = "word_structured"):
    for sylls_stream in sample_syllable_randomization(lexicon, max_tries=max_tries_randomize, tp_mode=tp_mode,
                                                      n_repetitions=n_repetitions):
        patterns = get_oscillation_patterns(len(lexicon[0].syllables))
        rhythmicity_indexes = compute_rhythmicity_index_sylls_stream(sylls_stream, patterns)
        if max_rhythmicity is None or (max(rhythmicity_indexes) <= max_rhythmicity):
            i_labels = enumerate(lexicon.info["syllables_info"]["syllable_feature_labels"])
            feature_labels = [f"phon_{i_phon+1}_{label}" for i_phon, labels in i_labels for label in labels]
        
            return Stream(
                id="_".join([syll.id for syll in sylls_stream[:5]]) + "..." + "_".join([syll.id for syll in sylls_stream[-5:]]),
                syllables=sylls_stream,
                info={
                    "rhythmicity_indexes": {k: float(v) for k, v in zip(feature_labels, rhythmicity_indexes)},
                    "lexicon": str(lexicon),
                    "lexicon_info": lexicon.info,
                    "stream_tp_mode": tp_mode,
                    **lexicon.info,
                }
            )


def make_stream_from_words(words: Register[str, Word],
                           n_words: int = 4,
                           max_word_overlap: int = 1,
                           max_lexicons: int = 10,
                           max_rhythmicity: float = 0.1,
                           max_tries_randomize: int = 10,
                           tp_mode: Literal["word_structured", "position_controlled", "random"] = "word_structured"
                           ) -> Optional[StreamType]:

    l_gen = make_lexicon_generator(words=words,
                                   n_words=n_words, max_overlap=max_word_overlap, max_yields=max_lexicons)

    for lex in l_gen:
        maybe_stream: Optional[StreamType] = make_stream_from_lexicon(lex, max_rhythmicity=max_rhythmicity,
                                                                      max_tries_randomize=max_tries_randomize,
                                                                      tp_mode=tp_mode)

        if maybe_stream:
            return maybe_stream


def get_stream_syllable_stats(stream: StreamType) -> Dict:
    d = {}
    for syllable in stream:
        if syllable.id in d:
            d[syllable.id] += 1
        else:
            d[syllable.id] = 1

    return d


def make_streams(
        lexicons: Optional[List[LexiconType]],
        max_rhythmicity: Optional[float] = None,
        stream_length: int = 15,
        max_tries_randomize: int = 10,
        tp_modes: tuple = ("random", "word_structured", "position_controlled"),
        require_all_tp_modes: bool = True
) -> RegisterType:
    """_summary_

    Args:
        lexicons (List[LexiconType]): A list of lexicons used as a basis for generatng the streams
        max_rhythmicity (Optional[float], optional): check rhythmicity and discard all streams that have at least one feature with higher PRI than this number. Defaults to None.
        stream_length (int, optional): how many syllables are in a stream in multiples of syllables in the lexicon. Defaults to 4.
        max_tries_randomize (int, optional): if max_rhythmicity is given and violated, how many times to try with a new randomization. Defaults to 10.
        tp_modes (tuple, optional): the ways (modes) in which to control for transition probabilities of syllables in the stream. Defaults to ("random", "word_structured", "position_controlled").
        require_all_tp_modes (bool, optional): all streams coming from the same lexicon will be discarded if not all their tp-modes have been found. Defaults to True.

    Returns:
        RegisterType: _description_
    """
    logger.info("Building streams from lexicons ...")

    streams = {}

    for i, lexicon in enumerate(lexicons):
        found_all_tp_modes = True
        new_streams = {}
        for tp_mode in tp_modes:

            maybe_stream: Optional[StreamType] = make_stream_from_lexicon(
                lexicon,
                max_rhythmicity=max_rhythmicity,
                max_tries_randomize=max_tries_randomize,
                n_repetitions=stream_length,
                tp_mode=tp_mode
            )

            if maybe_stream:
                new_streams[f"{''.join(word.id for word in lexicon)}_{tp_mode}"] = maybe_stream
            else:
                found_all_tp_modes = False
                logger.warning(f"Stream not found for lexicon {i} with tp_mode {tp_mode}.")
                break

        if found_all_tp_modes or (require_all_tp_modes == False):
            streams.update(new_streams)

    streams_reg = Register(**streams)

    streams_reg.info = {
        "tp_modes": tp_modes,
        "max_rhythmicity": max_rhythmicity,
        "max_tries_randomize": max_tries_randomize,
        "stream_length": stream_length,
        "require_all_tp_modes": require_all_tp_modes
    }

    return streams_reg



from functools import partial
import logging
import os.path
from copy import copy
from os import PathLike
from typing import Iterable, Dict, Union, Optional

import numpy as np
from scipy import stats

from alparc.types.base_types import Register, RegisterType
from alparc.types.phoneme import Phoneme
from alparc.types.syllable import Syllable
from alparc.types.word import Word

from alparc.io import read_phoneme_corpus, read_bigrams, read_trigrams, IPA_SEG_DEFAULT_PATH, IPA_BIGRAMS_DEFAULT_PATH, \
    IPA_TRIGRAMS_DEFAULT_PATH


logger = logging.getLogger(__name__)


def filter_uniform_syllables(syllables: Register[str, Syllable], alpha: float = 0.05):
    logger.info("Filter uniformly distributed syllables.")
    freqs = [s.info["freq"] for s in syllables]
    p_vals_uniform = stats.uniform.sf(abs(stats.zscore(np.log(freqs))))
    return Register({k: v for i, (k, v) in enumerate(syllables.items()) if p_vals_uniform[i] > alpha},
                    _info=copy(syllables.info))


def filter_common_phoneme_syllables(syllables, lang="deu"):
    logger.info("Filter syllables with common/native phonemes.")
    native_phonemes = read_phoneme_corpus(lang=lang)

    def is_native(syllable):
        return all((phoneme.id in native_phonemes) for phoneme in syllable)

    return Register({
        syllable.id: syllable for syllable in syllables if is_native(syllable)
    }, _info=copy(syllables.info))


def phoneme_is_common_at(phoneme: Phoneme, position: int = 0, p_threshold: float = 0.05):
    assert "word_position_prob" in phoneme.info.keys(), (
        "To check for phoneme position probability, your phonemes need the 'word_position_prob' info key. " 
        "Before creating syllables and words from your phonemes run `phonemes = phonemes.intersection(read_phoneme_corpus())`."
    )

    return phoneme.info["word_position_prob"].get(position, 0) >= p_threshold


def filter_common_phonemes_at_all_positions(word: Word, p_threshold: float = 0.05) -> bool:
    phonemes = [phoneme for syllable in word for phoneme in syllable]
    return all(phoneme_is_common_at(ph, position, p_threshold=p_threshold) for position, ph in enumerate(phonemes))

def filter_common_phonemes_at_position(word: Word, position, p_threshold: float = 0.05) -> bool:
    phonemes = [phoneme for syllable in word for phoneme in syllable]
    return phoneme_is_common_at(phonemes[position], position, p_threshold=p_threshold)


def filter_common_phoneme_words(words: RegisterType, position: Optional[int] = None, p_threshold: float = 0.05, 
                                ipa_seg_path: Union[str, PathLike] = IPA_SEG_DEFAULT_PATH):
    logger.info("Exclude words with low (onset) syllable probability.")

    if position is None:
        return words.filter(filter_common_phonemes_at_all_positions, p_threshold=p_threshold)
    else:
        return words.filter(filter_common_phonemes_at_position, position=position, p_threshold=p_threshold)


def check_bigram_stats(word: Word, valid_bigrams: Register[str, Syllable]):
    phonemes = [phon for syllable in word for phon in syllable]

    for phon_1, phon_2 in zip(phonemes[:-1], phonemes[1:]):
        if "".join([phon_1.id, phon_2.id]) not in valid_bigrams:
            return False

    return True


def check_trigram_stats(word: Word, valid_trigrams: Register[str, Syllable]):
    phonemes = [phon for syllable in word for phon in syllable]

    for phon_1, phon_2, phon_3 in zip(phonemes[:-2], phonemes[1:-1], phonemes[2:]):
        if "".join([phon_1.id, phon_2.id, phon_3.id]) not in valid_trigrams:
            return False

    return True


def filter_bigrams(words: RegisterType,
                   bigrams_path: Optional[Union[str, PathLike]] = IPA_BIGRAMS_DEFAULT_PATH,
                   p_val: float = None) -> Register[str, Word]:
    logger.info("Select words with uniform bigram and non-zero trigram log-probability of occurrence in the corpus.")

    words = copy(words)  # ?

    words.info.update({"bigram_pval": p_val})

    assert os.path.exists(bigrams_path), "Bigram control requires valid path to bigrams file"

    bigrams = read_bigrams(bigrams_path)

    if p_val is not None:
        bigrams = bigrams.filter(lambda bigram: bigram.info["p_unif"] > p_val)

    words.filter(check_bigram_stats, valid_bigrams=bigrams)

    return words


def filter_trigrams(words: RegisterType,
                    trigrams_path: Optional[Union[str, PathLike]] = IPA_TRIGRAMS_DEFAULT_PATH,
                    p_val: float = None) -> Register[str, Word]:
    logger.info("Select words with uniform bigram and non-zero trigram log-probability of occurrence in the corpus.")

    words = copy(words)  # ?

    words.info.update({"trigram_pval": p_val})

    assert os.path.exists(trigrams_path), "Trigram control requires valid path to trigrams file"
    trigrams = read_trigrams(trigrams_path)

    if p_val is not None:
        trigrams = trigrams.filter(lambda trigram: trigram.info["p_unif"] > p_val)
        
    words.filter(check_trigram_stats, valid_trigrams=trigrams)

    return words



import numpy as np
import random


def get_oscillation_patterns(lag):
    kernel = [1] + [0] * (lag - 1) + [1] + [0] * (lag - 1)
    return [list(np.roll(kernel, i)) for i in range(lag)]

def set_seed(seed: int):
    random.seed(seed)
    np.random.seed(seed)



from copy import copy
import itertools
import logging
from typing import TypeVar, List, Dict, Any

import numpy as np
from pydantic import BaseModel
from tqdm import tqdm

from alparc.types.base_types import Register, Element, RegisterType
from alparc.types.phoneme import PHONEME_FEATURE_LABELS, TypePhonemeFeatureLabels
from alparc.types.syllable import Syllable, SyllableType
from alparc.types.word import Word, WordType

from alparc.controls.common import *
from alparc.controls.filter import filter_bigrams, filter_common_phoneme_words, filter_trigrams


logger = logging.getLogger(__name__)


def check_syll_feature_overlap(syllables):
    all_feats = [feat for syll in syllables for phon_feats in syll.info["phonotactic_features"] for feat in phon_feats]
    return len(all_feats) == len(set(all_feats))


def generate_subset_syllables(syllables, lookback_syllables):
    subset = []
    for new_syll in syllables:
        syll_test_set = lookback_syllables + [new_syll]
        if check_syll_feature_overlap(syll_test_set):
            subset.append(new_syll)

    return subset


def word_overlap_matrix(
        words: Register[str, Word], 
        lag_of_interest: int = 1,
        control_features: List[TypePhonemeFeatureLabels] = PHONEME_FEATURE_LABELS):
    n_words = len(words)
    n_sylls_per_word = len(words[0].syllables)

    oscillation_patterns = get_oscillation_patterns(lag=(n_sylls_per_word*lag_of_interest))

    overlap = np.zeros([n_words, n_words], dtype=int)

    feature_indexes = []
    features = [feature for phoneme_features in words.info["syllables_info"]["syllable_feature_labels"] for feature in phoneme_features]
    for i, feature in enumerate(features):
        if feature in control_features:
            feature_indexes.append(i)

    for i1, i2 in list(itertools.product(range(n_words), range(n_words))):
        word_pair_features = [f1 + f2 for f1, f2 in zip(words[i1].info["binary_features"],
                                                        words[i2].info["binary_features"])]

        for i, word_pair_feature in enumerate(word_pair_features):
            if (word_pair_feature in oscillation_patterns) and i in feature_indexes:
                overlap[i1, i2] += 1

    return overlap


def generate_feature_words(syllables, iter_tries, n_syllables, n_look_back, phonotactic_control, progress_bar, n_words):
    words = {}

    if progress_bar:
        pbar = tqdm(total=n_words)

    for _ in iter_tries:
        sylls = []
        for _ in range(n_syllables):
            sub = list(filter(lambda x: x.id not in sylls, syllables))
            if phonotactic_control:
                sub = generate_subset_syllables(sub, sylls[-n_look_back:])
            if sub:
                new_rand_valid_syllable = random.sample(sub, 1)[0]
                sylls.append(new_rand_valid_syllable)

        if len(sylls) == n_syllables:
            word_id = "".join(s.id for s in sylls)
            if word_id not in words:
                word_features = list(list(tup) for tup in zip(*[s.info["binary_features"] for s in sylls]))
                words[word_id] = Word(id=word_id, info={"binary_features": word_features}, syllables=sylls)
                if progress_bar:
                    pbar.update(1)

        if len(words) == n_words:
            logging.info(f"Done: Found {n_words} words.")
            break
    
    return Register(words, _info={"n_syllables_per_word": n_syllables, "n_look_back": n_look_back, "phonotactic_control": phonotactic_control})



def make_words(syllables: RegisterType,
               num_syllables=3,
               bigram_control=True,
               bigram_alpha=None,
               trigram_control=True,
               trigram_alpha=None,
               positional_control=True,
               positional_control_position=None,
               position_alpha=0,
               phonotactic_control=True,
               n_look_back=2,
               n_words=10_000,
               max_tries=100_000,
               progress_bar: bool = True,
               lang="deu"
               ) -> RegisterType:
    """_summary_

    Args:
        syllables (RegisterType): The Register of syllables to use as a basis for word generation
        num_syllables (int, optional): how many syllables are in a word. Defaults to 3.
        bigram_control (bool, optional): apply statistical control on the bigram level. Defaults to True.
        bigram_alpha (_type_, optional): which p-value to assume for bigram control. Defaults to None.
        trigram_control (bool, optional): apply statistical control on the trigram level. Defaults to True.
        trigram_alpha (_type_, optional): which p-value to assume for trigram control. Defaults to None.
        positional_control (bool, optional): control phoneme positions in words to be likely given the language. Defaults to True.
        positional_control_position (int, optional): At which position to control phoneme likelihood (None means all). Defaults to None.
        position_alpha (float, optional): probability throshold for positional control. Defaults to 0.
        phonotactic_control (bool, optional): control each syllabel for minimum feature overlap with previous syllables. Defaults to True.
        n_look_back (int, optional): how far to look back in the feature overlap control of the syllables. Defaults to 2.
        n_words (_type_, optional): how many words to generate. Defaults to 10_000.
        max_tries (_type_, optional): how often to attemt to add a word to the Register, before the function gives up. Defaults to 100_000.
        progress_bar (bool, optional): print a progress bar based on 'n_words'. Defaults to True.

    Returns:
        RegisterType: The Register of words.
    """

    iter_tries = range(max_tries)

    words_register = generate_feature_words(syllables, iter_tries, num_syllables, n_look_back, phonotactic_control, progress_bar, n_words)

    words_register.info["syllables_info"] = copy(syllables.info)

    if bigram_control and lang=="deu":
        logger.info("bigram control...")
        words_register = filter_bigrams(words_register, p_val=bigram_alpha)

    if trigram_control and lang=="deu":
        logger.info("trigram control...")
        words_register = filter_trigrams(words_register, p_val=trigram_alpha)

    if positional_control and lang=="deu":
        logger.info("positional control...")
        words_register = filter_common_phoneme_words(words_register, p_threshold=position_alpha, position=positional_control_position)

    return words_register



from .io import load_phonemes, load_syllables, load_words, load_lexicons, load_streams, export_speech_synthesizer

from .controls.common import set_seed

from .types.base_types import Register, RegisterType, Element
from .types.phoneme import Phoneme, PhonemeType
from .types.syllable import Syllable, SyllableType
from .types.word import Word, WordType
from .types.lexicon import Lexicon, LexiconType
from .types.stream import Stream, StreamType

from .core.syllable import make_syllables
from .core.word import make_words
from .core.lexicon import make_lexicons
from .core.stream import make_streams

from .eval import to_lexicon, to_stream, to_word



import json
from os import PathLike
from typing import TypeVar, List, Dict, Any, Union

from pydantic import BaseModel

from alparc.types.base_types import Register, Element, RegisterType
from alparc.types.syllable import Syllable, SyllableType


StreamType = TypeVar("StreamType", bound="Stream")


class Stream(Element, BaseModel):
    id: str
    syllables: List[SyllableType]
    info: Dict[str, Any]

    def get_elements(self):
        return self.syllables
        
    def __str__(self):
        return "_".join([syll.id for syll in self.syllables[:5]]) + "..." + "_".join([syll.id for syll in self.syllables[-5:]])
    
    def save(self, path: Union[str, PathLike] = None):
        if path is None:
            path = f"stream.json"

        if isinstance(path, str) and not path.endswith(".json"):
            path = path + ".json"

        with open(path, "w", encoding='utf-8') as file:
            json.dump(self, file,
                      default=lambda o: o.model_dump(), sort_keys=False, ensure_ascii=False)



import json
from abc import ABC, abstractmethod
from collections import OrderedDict
from copy import copy
from os import PathLike
from typing import Dict, Any, Type, TypeVar, Union

import pandas as pd

from alparc.controls.common import *
from alparc.types.elements import Element

RegisterType = TypeVar("RegisterType", bound="Register")


class Register(OrderedDict):
    MAX_PRINT_ELEMENTS = 10
    INFO_KEY = "_info"

    def __init__(self, other=(), /, **kwargs):
        if self.INFO_KEY in kwargs:
            self.info = kwargs[self.INFO_KEY]
            del kwargs[self.INFO_KEY]
        else:
            self.info = {}

        super().__init__(other, **kwargs)

    def __contains__(self, item: Union[str, Element]):
        if isinstance(item, str):
            return item in self.keys()
        elif isinstance(item, Element):
            return item.id in self.keys()
        else:
            raise ValueError("item type unknown")

    @property
    def info(self):
        return self._info

    @info.setter
    def info(self, value: Dict):
        self._info = value

    def __iter__(self):
        return iter(self.values())

    def __getitem__(self, item):
        if isinstance(item, (int, slice)):
            return list(self.values())[item]

        return super().__getitem__(item)

    def __str__(self):
        li = list(self.keys())
        n_elements = len(li)

        s = "|".join(li[:self.MAX_PRINT_ELEMENTS])

        if n_elements > self.MAX_PRINT_ELEMENTS:
            s += "|..."

            s += f" ({n_elements} elements total)"

        return s

    def append(self, obj: Element):
        self[str(obj)] = obj

    def get_subset(self, size: int) -> RegisterType:
        """Create a new Register as a random subset of this one"""
        if size >= len(self):
            return self

        keys = set()

        for _ in range(size):
            keys.add(random.choice(list(self.keys() - keys)))

        return self.new_from_dict({key: self[key] for key in keys})

    def get_self_with_info_key(self):
        d = copy(self)
        d.update({self.INFO_KEY: self.info})
        return d

    def to_json(self):
        return json.dumps(self.get_self_with_info_key(), default=lambda o: o.model_dump(),
                          sort_keys=False, ensure_ascii=False)

    def save(self, path: Union[str, PathLike] = None):
        if path is None:
            path = f"{self[0].__class__.__name__.lower()}s.json"

        if isinstance(path, str) and not path.endswith(".json"):
            path = path + ".json"

        with open(path, "w", encoding='utf-8') as file:
            json.dump(self.get_self_with_info_key(), file,
                      default=lambda o: o.model_dump(), sort_keys=False, ensure_ascii=False)

    def intersection(
            self,
            other: RegisterType
    ) -> RegisterType:
        """
        Select Elements that are in both registers and merge the data
        :param other:
        :return:
        """

        def merge_infos(element_1, element_2):
            element_new = copy(element_1)
            element_new.info.update(element_2.info)
            return element_new

        new_register = self.new_from_dict({
            key: merge_infos(element, other[key]) for key, element in self.items() if key in other
        })

        new_register.info.update(other.info)

        return new_register

    def new_from_dict(self, dictionary: dict) -> RegisterType:
        new_info = copy(self.info)
        return Register(dictionary, _info=new_info)
    
    def empty_like(self):
        return Register({}, _info=copy(self.info))

    def flatten(self) -> RegisterType:
        reg = self.empty_like()
        for element in self:
            for sub_element in element.get_elements():
                reg.append(sub_element)
        return reg
    
    def filter(self, func, *args, **kwargs):
        reg = self.empty_like()
        for element in self:
            if func(element, *args, **kwargs):
                reg.append(element)
        return reg



import csv
import json
import logging
import os
import pathlib
from importlib import resources as importlib_resources
from os import PathLike
from typing import Iterable, Dict, Union, List, Type, Optional, Literal
from functools import partial
from copy import copy

import numpy as np

from alparc.io import load_phonemes
from alparc.types.base_types import Register, RegisterType

from alparc.types.phoneme import PHONEME_FEATURE_LABELS, Phoneme
from alparc.types.syllable import Syllable
from alparc.types.word import Word
from alparc.types.lexicon import LexiconType
from alparc.types.stream import Stream

from alparc.core.syllable import LABELS_C, LABELS_V, syllable_from_phonemes
from alparc.core.word import Word, word_overlap_matrix
from alparc.core.stream import compute_rhythmicity_index_sylls_stream, get_oscillation_patterns

ALL_DEFAULT_PHONEMES = load_phonemes(lang=None)
SYLLABLE_FEAT_LABELS = [LABELS_C] + [LABELS_V]

def to_syllable(syllable, syllable_type="cv"):
    says_cv = (syllable_type == "cv")
    says_cV = (syllable_type == "cV")

    is_cV = (len(syllable) == 3) and syllable.endswith("ː")
    is_cv = (len(syllable) == 2) and not syllable.endswith("ː")
    is_diphthong = (len(syllable) == 3 and not syllable.endswith("ː"))

    if not any([says_cv and is_cv, says_cV and is_cV, says_cv and is_diphthong]):
        raise ValueError(f"The syllable type '{syllable_type}' does not match for syllable {syllable}. "
                          "The types 'cv' (single character consonant + short vowel) and 'cV' (single "
                          "character consonant + long vowel) are supported. Short vowel can also be a diphthong. " 
                          "All syllables must be of that type.")
    
    if is_diphthong:
        syllable_obj = syllable_from_phonemes(ALL_DEFAULT_PHONEMES, syllable[:2], SYLLABLE_FEAT_LABELS)
        syllable_obj.id = syllable
        return syllable_obj
    
    if is_cV:
        return syllable_from_phonemes(ALL_DEFAULT_PHONEMES, [syllable[:-2], syllable[-2:]], SYLLABLE_FEAT_LABELS)
        
    return syllable_from_phonemes(ALL_DEFAULT_PHONEMES, syllable, SYLLABLE_FEAT_LABELS)

def to_word(word, syllable_type="cv"):
    to_syllable_partial = partial(to_syllable, syllable_type=syllable_type)
    syllables_list = list(map(to_syllable_partial, word))
    word_id = "".join(s.id for s in syllables_list)
    word_features = list(list(tup) for tup in zip(*[s.info["binary_features"] for s in syllables_list]))
    return Word(id=word_id, info={"binary_features": word_features}, syllables=syllables_list)

def to_lexicon(lexicon, syllable_type="cv"):
    to_word_partial = partial(to_word, syllable_type=syllable_type)
    word_objs_list = list(map(to_word_partial, lexicon))
    lexicon = Register({w.id:  w for w in word_objs_list})
    lexicon.info.update({"syllables_info": {"syllable_feature_labels": [LABELS_C, LABELS_V],  "syllable_type": syllable_type}})
    overlap = word_overlap_matrix(lexicon)
    lexicon.info["cumulative_feature_repetitiveness"] = int(np.triu(overlap, 1).sum())
    lexicon.info["max_pairwise_feature_repetitiveness"] = int(np.triu(overlap, 1).max())
    return lexicon

def to_stream(stream: list, syllable_type: str = "cv", lag_of_interest: int = 3):
    """Generate stream object from stream text file.

    Args:
        stream (list): stream as list of strings
        syllable_type (str, optional): phoneme pattern for syllables. Defaults to "cv".
        lag_of_interest (int, optional): Oscillation patterns for feature matching. 
                                         Usually equal to the number of syllables in a word. Defaults to 3.

    Raises:
        ValueError: Syllable type not supported.

    Returns:
        Stream: The parsed stream.
    """
    if syllable_type not in ["cv", "cV"]:
        raise ValueError(f"Syllable type {syllable_type} not supported, only one of ['cv', 'cV'].")
    to_syllable_partial = partial(to_syllable, syllable_type=syllable_type)
    syllables_list = list(map(to_syllable_partial, stream))
    stream_id = "".join(s.id for s in syllables_list)

    stream = Stream(id=stream_id, info={"lexicon": None}, syllables=syllables_list)

    patterns = get_oscillation_patterns(lag_of_interest)
    rhythmicity_indexes = compute_rhythmicity_index_sylls_stream(stream, patterns)
    i_labels = enumerate(SYLLABLE_FEAT_LABELS)
    feature_labels = [f"phon_{i_phon+1}_{label}" for i_phon, labels in i_labels for label in labels]
    stream.info.update({"rhythmicity_indexes": {k: float(v) for k, v in zip(feature_labels, rhythmicity_indexes)}})

    return stream



from typing import Literal, get_args, TypeVar, Dict, Any

from pydantic import BaseModel

from alparc.types.base_types import Element

TypePhonemeFeatureLabels = Literal[
    "syl", "son", "cons", "cont", "delrel", "lat", "nas", "strid", "voi", "sg", "cg", "ant", "cor", "distr", "lab",
    "hi", "lo", "back", "round", "tense", "long"
]
PHONEME_FEATURE_LABELS = list(get_args(TypePhonemeFeatureLabels))
PhonemeType = TypeVar("PhonemeType", bound="Phoneme")


class Phoneme(Element, BaseModel):
    id: str
    info: Dict[str, Any]

    def get_elements(self):
        return []

    def get_feature_symbol(self, label: TypePhonemeFeatureLabels):
        return self.info["features"][PHONEME_FEATURE_LABELS.index(label)]

    def get_binary_feature(self, label: TypePhonemeFeatureLabels):
        return self.get_feature_symbol(label) == "+"



from typing import List, Literal, Optional, Union, TypeVar, Dict, Any

from pydantic import BaseModel

from alparc.types.phoneme import Phoneme
from alparc.types.base_types import Element

LABELS_C = ['son', 'back', 'hi', 'lab', 'cor', 'cont', 'lat', 'nas', 'voi']
LABELS_V = ['back', 'hi', 'lo', 'lab', 'tense', 'long']
N_FEAT = len(LABELS_C) + len(LABELS_V)  # 14

SyllableType = TypeVar("SyllableType", bound="Syllable")


class Syllable(Element, BaseModel):
    id: str
    phonemes: List[Phoneme]
    info: Dict[str, Any]

    def get_elements(self):
        return self.phonemes


import json
from abc import ABC, abstractmethod
from collections import OrderedDict
from copy import copy
from os import PathLike
from typing import Dict, Any, TypeVar, Union

from alparc.controls.common import *


class Register(OrderedDict):
    MAX_PRINT_ELEMENTS = 10
    INFO_KEY = "_info"

    def __init__(self, other=(), /, **kwargs):
        if self.INFO_KEY in kwargs:
            self.info = kwargs[self.INFO_KEY]
            del kwargs[self.INFO_KEY]
        else:
            self.info = {}

        super().__init__(other, **kwargs)

    def __contains__(self, item: Union[str, Element]):
        if isinstance(item, str):
            return item in self.keys()
        elif isinstance(item, Element):
            return item.id in self.keys()
        else:
            raise ValueError("item type unknown")

    @property
    def info(self):
        return self._info

    @info.setter
    def info(self, value: Dict):
        self._info = value

    def __iter__(self):
        return iter(self.values())

    def __getitem__(self, item):
        if isinstance(item, (int, slice)):
            return list(self.values())[item]

        return super().__getitem__(item)

    def __str__(self):
        li = list(self.keys())
        n_elements = len(li)

        s = "|".join(li[:self.MAX_PRINT_ELEMENTS])

        if n_elements > self.MAX_PRINT_ELEMENTS:
            s += "|..."

            s += f" ({n_elements} elements total)"

        return s

    def append(self, obj: Element):
        self[str(obj)] = obj

    def get_subset(self, size: int) -> RegisterType:
        """Create a new Register as a random subset of this one"""
        if size >= len(self):
            return self

        keys = set()

        for _ in range(size):
            keys.add(random.choice(list(self.keys() - keys)))

        return self.new_from_dict({key: self[key] for key in keys})

    def get_self_with_info_key(self):
        d = copy(self)
        d.update({self.INFO_KEY: self.info})
        return d

    def to_json(self):
        return json.dumps(self.get_self_with_info_key(), default=lambda o: o.model_dump(),
                          sort_keys=False, ensure_ascii=False)

    def save(self, path: Union[str, PathLike] = None):
        if path is None:
            path = f"{self[0].__class__.__name__.lower()}s.json"

        if isinstance(path, str) and not path.endswith(".json"):
            path = path + ".json"

        with open(path, "w", encoding='utf-8') as file:
            json.dump(self.get_self_with_info_key(), file,
                      default=lambda o: o.model_dump(), sort_keys=False, ensure_ascii=False)

    def intersection(
            self,
            other: RegisterType
    ) -> RegisterType:
        """
        Select Elements that are in both registers and merge the data
        :param other:
        :return:
        """

        def merge_infos(element_1, element_2):
            element_new = copy(element_1)
            element_new.info.update(element_2.info)
            return element_new

        new_register = self.new_from_dict({
            key: merge_infos(element, other[key]) for key, element in self.items() if key in other
        })

        new_register.info.update(other.info)

        return new_register

    def new_from_dict(self, dictionary: dict) -> RegisterType:
        new_info = copy(self.info)
        return Register(dictionary, _info=new_info)
    
    def empty_like(self):
        return Register({}, _info=copy(self.info))

    def flatten(self) -> RegisterType:
        reg = self.empty_like()
        for element in self:
            for sub_element in element.get_elements():
                reg.append(sub_element)
        return reg
    
    def filter(self, func, *args, **kwargs):
        reg = self.empty_like()
        for element in self:
            if func(element, *args, **kwargs):
                reg.append(element)
        return reg

class PhonemeRegister(Register):
    pass

class SyllableRegister(Register):
    pass

class WordRegister(Register):
    pass

class Lexicon(WordRegister):
    pass

class StreamRegister(Register):
    pass

